/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: irc.c 148 2004-05-24 22:13:07Z pasi $ */


#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include <assert.h>

#include "spamserv.h"
#include "structs.h"
#include "server.h"
#include "irc.h"
#include "users.h"
#include "ui.h"
#include "log.h"
#include "motd.h"
#include "magic.h"
#include "formats.h"
#include "pseudoclient.h"
#include "util.h"

extern struct cmd cmdtab[];


irc_user *i_alloc_irc_user(void)
{
	return (irc_user *)g_malloc0(sizeof(irc_user));
}

void i_send_globops(char *fmt, ...)
{
	char *buf;
	va_list ap;

	va_start(ap, fmt);
	buf = g_strdup_vprintf(fmt, ap);
	send_raw("GLOBOPS :%s", buf);
	g_free(buf);
	va_end(ap);
}

void i_send_notice(irc_user *user, char *fmt, ...)
{
	char *buf;
	va_list ap;

	g_return_if_fail(user != NULL);
	
	va_start(ap, fmt);
	buf = g_strdup_vprintf(fmt, ap);
	send_raw(":%s NOTICE %s :%s", "SpamServ", 
		 user->nick, buf);
	g_free(buf);
	va_end(ap);
}

/* This takes a char * argument instead of irc_user because private
 * messages should only be sent to channels. When we reply
 * to users, use notices (RFC1569). */
void i_send_privmsg(char *target, char *fmt, ...)
{
        char *buf;
        va_list ap;

	g_return_if_fail(target != NULL);	
	
        va_start(ap, fmt);
        buf = g_strdup_vprintf(fmt, ap);
        send_raw(":%s PRIVMSG %s :%s", "SpamServ",
                 target, buf);
        g_free(buf);
        va_end(ap);
}

void i_send_whois(irc_user *target, irc_user *data, GString *chanlist)
{
	g_return_if_fail(target != NULL);
	g_return_if_fail(data != NULL);
	
	(void)s_fill_user_by_nick(target->nick, target);
	send_raw("311 %s %s %s %s * :%s", target->nick,
		 data->nick, data->username, data->host,
		 data->gcos);
	if (IS_SERVICESADMIN(target)) {
		if (chanlist)
			send_raw("319 %s %s :%s", target->nick, data->nick,
				 chanlist->str);
	} else 
		send_raw("319 %s %s :&channel_list_hidden", target->nick,
			 data->nick);
	send_raw("312 %s %s %s :%s", target->nick, data->nick,
		 conf.servername, conf.serverdesc);
	send_raw("317 %s %s %ld %ld :second idle, signon time",
		 target->nick, data->nick, time(NULL) - status.started,
		 status.started);
	send_raw("318 %s %s :End of /WHOIS list.", target->nick,
		 data->nick);
}

void i_send_nick(irc_user *user)
{
	g_return_if_fail(user != NULL);

	send_raw("NICK %s 1 %ld +w %s %s %s 0 2130706433 :%s",
			user->nick, time(NULL), user->username,
			user->host, conf.servername, user->gcos);
}

void i_send_sjoin(irc_user *user, char *channel)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(channel != NULL);

	send_raw(":%s SJOIN %lu %s", user->nick, time(NULL), channel);
}

void i_send_part(irc_user *user, char *channel)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(channel != NULL);
	
	send_raw(":%s PART %s", user->nick, channel);
}

void i_autokill_user(irc_user *user, int mins, char *reason)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(reason != NULL);
	
	send_raw("AKILL %s * %d %s %lu :%s", user->host,
			mins, "SpamServ", time(NULL), reason);
}

void i_kill_user(irc_user *user, char *reason)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(reason != NULL);
	
	send_raw(":%s KILL %s :%s!%s!%s (%s)", "SpamServ", user->nick,
			"SpamServ", conf.servername, "SpamServ",
			reason);
}

void i_send_quit(irc_user *user, char *reason)
{
	g_return_if_fail(user != NULL);
	g_return_if_fail(reason != NULL);
	
	send_raw(":%s QUIT :%s", user->nick, reason);
}

void i_parse_stats(irc_msg *msg)
{
	struct cmd *p_cmd;
	
	if (msg->msgtoks[0] == NULL)
		return;
	switch (msg->msgtoks[0][0]) {
	case 'c':
		if (is_privileged_oper(msg->user.nick))
			send_raw(FORMAT_STATS_c_O, msg->user.nick,
				 "*", conf.remoteserver,
				 conf.remoteport);
		else
			send_raw(FORMAT_STATS_c, msg->user.nick);
		break;
	case 'm':
		for (p_cmd = cmdtab; p_cmd->cmd; p_cmd++)
			if (p_cmd->usagecount)
				 send_raw(FORMAT_STATS_m, msg->user.nick,
					  p_cmd->cmd, p_cmd->usagecount);
		break;
	case 'x':
		if (!is_privileged_oper(msg->user.nick))
			break;
		send_raw("219 %s Parsed private messages: %lu", msg->user.nick, status.stats.parsed_privmsgs);
		send_raw("219 %s Autokills: %lu", msg->user.nick, status.stats.set_autokills);
		send_raw("219 %s Kills: %lu", msg->user.nick, status.stats.killed_users);
		send_raw("219 %s Matches: %lu", msg->user.nick, status.stats.matched_spams);
		send_raw("219 %s Forked: %d", msg->user.nick, status.forked);
		send_raw("219 %s Started: %lu", msg->user.nick, status.started);
		send_raw("219 %s Database FD: %d", msg->user.nick, status.dbconnection);
		send_raw("219 %s Quitting: %d", msg->user.nick, status.quitting);
		send_raw("219 %s Remote server: <hidden>", msg->user.nick);
		send_raw("219 %s Remote port: %d", msg->user.nick, conf.remoteport);
		send_raw("219 %s Server desc: %s", msg->user.nick, conf.serverdesc);
		send_raw("219 %s Autokill len: %d", msg->user.nick, conf.akilllen);
		send_raw("219 %s Report chan: %s", msg->user.nick, conf.reportchan);
		send_raw("219 %s Info chan: %s", msg->user.nick, conf.infochan);
		send_raw("219 %s DB:", msg->user.nick);
		send_raw("219 %s User: <hidden>", msg->user.nick);
		send_raw("219 %s Pass: <hidden>", msg->user.nick);
		send_raw("219 %s Database: %s", msg->user.nick, conf.db.database);
		send_raw("219 %s Host: %s", msg->user.nick, conf.db.host);
		send_raw("219 %s Port: %d", msg->user.nick, conf.db.port);
		
	}
	send_raw(FORMAT_STATS_END, msg->user.nick,
		 msg->msgtoks[0][0]);
}

void i_parse_info(irc_msg *msg)
{
	send_raw(FORMAT_INFO_START, msg->user.nick);
	send_raw(FORMAT_INFO, msg->user.nick,
			"Written by Pasi Hirvonen and David Page");
	send_raw(FORMAT_INFO, msg->user.nick, "-");
	send_raw(FORMAT_INFO, msg->user.nick, "Based on the original work by: ");
	send_raw(FORMAT_INFO, msg->user.nick, "Matthew Stephanson and Pasi Hirvonen");
	send_raw(FORMAT_INFO, msg->user.nick, "Special thanks to: ");
	send_raw(FORMAT_INFO, msg->user.nick, " - Devon (Devon Bleak)      - Coding tips etc.");
	send_raw(FORMAT_INFO, msg->user.nick, " - Kobi_S                   - Fixing!");
	send_raw(FORMAT_INFO, msg->user.nick, " - gcc                      - Harassment.");
	send_raw(FORMAT_INFO, msg->user.nick, " - Robyn                    - Ideas!");
	send_raw(FORMAT_INFO, msg->user.nick, " - terminal (Ryan Pugatch)  - Beta testing");
	send_raw(FORMAT_INFO, msg->user.nick, " - derfy (Freddie Agricola) - Segfaulting (aka Beta testing)");
	send_raw(FORMAT_INFO, msg->user.nick, " - IRCtoo people!");
	send_raw(FORMAT_INFO, msg->user.nick, "If you're not listed and you should be, let us know.");
	send_raw(FORMAT_INFO_END);
}

void i_parse_motd(irc_msg *msg)
{
	GSList *gs;
	gs=s_cache_motd();
	send_raw(FORMAT_MOTD_START, msg->user.nick, conf.servername);
	g_slist_foreach(gs, foreach_motd_line, msg->user.nick);
	send_raw(FORMAT_MOTD_END, msg->user.nick);
}

void foreach_motd_line(gpointer data, gpointer nickname)
{
	send_raw("372 %s :- %s", nickname, data);
}

void i_parse_nick(irc_msg *msg)
{
	irc_user *u;
	
	if (msg->msgtoks[0] == NULL)
		return;
	u = s_find_user(msg->user.nick);
	/* do we already have this user? */
	if (u != NULL) {
		s_log(LOG_DEBUG, "Nick change, %s -> %s.", msg->user.nick,
		      msg->msgtoks[0]);
		g_strlcpy(u->nick, msg->msgtoks[0], MAXNICKLEN);
	} else {
		s_parse_new_user(msg);
	}
}

void i_parse_quit(irc_msg *msg)
{
	s_del_user(msg->user.nick);
}

void i_parse_kick(irc_msg *msg)
{
	if (msg->msgtoks[0] == NULL || msg->msgtoks[1] == NULL)
		return;
	s_log(LOG_DEBUG, "Removing pseudochan %s", msg->msgtoks[0]);
	s_del_pseudochan_by_kick(msg->msgtoks[1], msg->msgtoks[0]);
}

void i_parse_version(irc_msg *msg)
{
	 send_raw("351 %s %s %s :%s", msg->user.nick, SPAMSERV_VERSION
			EXTRA_VERSION, conf.servername, SPAMSERV_RELNAME);
}

void i_parse_ping(irc_msg *msg)
{
	char *s;

	/* Okay, okay. Little bit hackey. Third ping means that we're
	 * synchronized, this allows us to do some extra magic. */
	if (status.synched < 4) {
		status.synched++;
		if (status.synched == 3)
			s_do_synch();
	} 
	s = g_strjoinv(" ", msg->msgtoks);
	send_raw("PONG %s", s);
	g_free(s);
}

void i_parse_admin(irc_msg *msg)
{
	send_raw("256 %s %s :Administrative info",
		 msg->user.nick,conf.servername);
	send_raw("257 %s :Somewhere in teh yankland", msg->user.nick);
	send_raw("258 %s :http://www.irctoo.net/", msg->user.nick);
	send_raw("259 %s :Contact irctoo@irctoo.net for questions",
		 msg->user.nick);
}

void i_parse_invite(irc_msg *msg)
{
	char *chan;
	
	if (msg->msgtoks[0] == NULL || msg->msgtoks[1] == NULL)
		return;
	chan = (msg->msgtoks[1][0] == ':') ?
		msg->msgtoks[1] + 1 : msg->msgtoks[1];
	/* we ignore the return value here because we shouldn't
	 * normally get malformed channel name here (if yes, just
	 * ignore it) and if the chan is already known,
	 * invite_pseudoclient() sends the warning. */
	(void)s_invite_pseudoclient(chan, msg->msgtoks[0], &msg->user);
}

void i_parse_kill(irc_msg *msg)
{
	pclient *p;
	p = s_find_pseudoclient(msg->msgtoks[0]);
	if (p == NULL)
		return;
	/* one of our pseudoclients have been killed. must be reintroduced,
	 * and the person that did it, sacrificed in our tribal rituals. */
	s_clean_pseudoclient(p, NULL);
	s_introduce_pseudoclient(p,NULL);
	
}

void i_parse_privmsg(irc_msg *msg)
{
	irc_user *u;
	
	if (msg->msgtoks[0] == NULL)
		return; 
	u = s_find_user(msg->user.nick);
	if (u == NULL) {
		s_log(LOG_DEBUG, "Umm private message from unknown user.");
		return;
	}
	msg->target = msg->msgtoks[0];
	/* this also includes a leathal assumption (it doesn't *crash*
	 * or anything if the message doesn't contain a :, but will
	 * behave incorrectly because private message with only
	 * one word doesn't have to contain :) */
	if ((msg->msg = strchr(msg->msg, ':')) == NULL)
		return; 
	if (*(msg->msg + 1) != '\0')
		msg->msg++;
	if (g_ascii_strcasecmp(msg->msgtoks[0], "spamserv") == 0)
		if (IS_OPER(u))
			s_parse_command(msg);
		else
			s_send_helpfile(&msg->user, "regularuser");
	else if (!IS_OPER(u) && !IS_REGISTERED(u))
		spam_str_do_spamchecks(msg);
}

void i_parse_whois(irc_msg *msg)
{
	/* :six WHOIS spamserv :spamserv
	 */
	if (msg->msgtoks[0])
		s_pseudoclient_parse_whois(&msg->user, msg->msgtoks[0]);
}

void i_parse_lusers(irc_msg *msg)
{
	unsigned long users, locals;
	
	if (msg->msgtoks[0] == NULL ||  msg->msgtoks[1] == NULL)
		return;

	locals = status.stats.pseudoclient_count;
	users = status.stats.global_users + locals;
	send_raw("351 %s :There are %d users and %d invisible on %s servers",
		 msg->user.nick, users, users, "n");
	send_raw("351 %s %d :IRC Operators online", msg->user.nick, 5);
	send_raw("351 %s :I have %d clients and 1 servers", msg->user.nick,
		 locals);
	send_raw("351 %s :Current local users: %d Max: %d", msg->user.nick,
		 locals, locals);
	send_raw("351 %s :Current global users: %d Max: %d", msg->user.nick,
		 users, users);
}

void i_parse_svsmode(irc_msg *msg)
{
	if (msg->msgtoks[0] == NULL || msg->msgtoks[1] == NULL ||
	    msg->msgtoks[2] == NULL)
		return;
	/* drop channel mode */
	if (msg->msgtoks[0][0] == '#')
		return;
	s_parse_mode_change(msg->msgtoks[0], msg->msgtoks[2]);
}

void i_parse_mode(irc_msg *msg)
{
	if (msg->msgtoks[0] == NULL ||  msg->msgtoks[1] == NULL)
		return;
	/* channel mode? drop */
	if (msg->msgtoks[0][0] == '#') {
		/* We ignore modes from ChanServ */
		if (msg->lasttok < 3
		    || (g_ascii_strcasecmp(msg->user.nick, "ChanServ") == 0))
			return;
		s_parse_channel_mode(msg->msgtoks[0], msg->msgtoks,
				     msg->lasttok);
		return;
	}
	s_parse_mode_change(msg->msgtoks[0], msg->msgtoks[1]);
}

void s_parse_channel_mode(char *channel, char **msgtoks, int numtoks) {
	int i, inserting, modessofar, devoice, deop;
	char *pseudoname, *s;

	if (msgtoks[1] == NULL || msgtoks[2] == NULL ||
	    msgtoks[3] == NULL)
		return;
	/* note to self: 2 = modes, 3+ = other stuff */

	devoice = deop = 0;
	pseudoname = NULL;
	/* are we 'adding' stuff or taking it away? */
	inserting = 0;
	/* a counter for the actual mode letters so far */
	modessofar = 0;
	for (i = 0; msgtoks[2][i] != '\0'; ) {
			if (msgtoks[2][i] == '+') {
			inserting = 1;
			i++;
		} else if (msgtoks[2][i] == '-') {
			inserting = 0;
			i++;
		} else if (!inserting) {
			modessofar++;
			i++;
		} else if (msgtoks[2][i] == 'v' || msgtoks[2][i] == 'o') {
			modessofar++;
			/* this shouldn't really be possible.. */
			if (numtoks < (modessofar + 2))
				return;
			/* we only need to look it up once, it won't
			 * obviously(?) change. */
			if (pseudoname == NULL)
				pseudoname = s_pseudo_at_chan(channel);
			/* this is actually only necessary for the last
			   token. */
			for (s =  msgtoks[2 + modessofar]; *s != '\0'; *s++)
				if (*s == '\r')
					 *s = '\0';
			if (g_ascii_strcasecmp(pseudoname,
						msgtoks[2 + modessofar]) == 0) {
				if (msgtoks[2][i] == 'v')
					devoice = 1;
				else
					deop = 1;
				break;
			}
			i++;
		} else 
			i++;
	}
	if (devoice)
		send_raw("MODE %s -v :%s", channel, pseudoname);
	if (deop)
		send_raw("MODE %s -o :%s", channel, pseudoname);
}
