/* Copyright (C) 2003, 2004 IRCtoo Network, Pasi Hirvonen <psh@iki.fi>
 *
 * Parts written by FJ (David). Based on the original work by Pasi Hirvonen
 * and Matthew Stephenson. Various others have contributed ideas and testing.*/

/* $Id: users.c 148 2004-05-24 22:13:07Z pasi $ */


#include <glib.h>
#include <string.h>
#include <stdio.h>

#include "structs.h"
#include "spamserv.h"
#include "irc.h"
#include "server.h"
#include "users.h"
#include "log.h"

GSList *users = NULL;

struct hostsrc_args {
	GArray *resultstore;
	char *searchstring;
};

int s_usercmp(gpointer, gpointer);
void s_clear_userentry(gpointer, gpointer);
int s_parse_userflags(char *, int);
/* src is an interesting way to abbreviate "search" ;) */
void s_hostsrc(gpointer, gpointer);

void s_add_user(char *nick, char *username, char *host, char *gcos,
		int user_flags)
{
	irc_user *u;

	u = i_alloc_irc_user();
	g_strlcpy(u->nick, nick, MAXNICKLEN);
	g_strlcpy(u->username, username, MAXUSERLEN);
	g_strlcpy(u->host, host, MAXHOSTLEN);
	g_strlcpy(u->gcos, gcos, MAXREALLEN);

	s_log(LOG_DEBUG, "Added user %s (%s@%s) [%s].", nick, username,
	      host, gcos);
	users = g_slist_append(users, u);
	u->user_flags = user_flags;
	status.stats.global_users++;
}

int s_usercmp(gpointer p1, gpointer p2)
{
	return g_ascii_strncasecmp(((irc_user *)p1)->nick, (char *)p2,
			MAXNICKLEN);
}

void s_hostsrc(gpointer data, gpointer args)
{
	irc_user *u;
	struct hostsrc_args *ha;

	ha = (struct hostsrc_args *)args;
	u = (irc_user *)data;
	
	/* BROKEN CODE, use and bobstay will eat your bollocks */
#if 0
	printf("XXX %s ja %s XXX\n", u->host, ha->searchstring);
	if (g_pattern_match_simple(ha->searchstring, u->host) == TRUE)
		ha->resultstore = g_array_append_vals(ha->resultstore, u, 1);
#endif
}

int is_privileged_sa(char *nick)
{
	irc_user *u;

	u = s_find_user(nick);
	if (u == NULL)
		return 0;
	return IS_SERVICESADMIN(u);
}

int is_privileged_oper(char *nick)
{
	irc_user *u;

	u = s_find_user(nick);
	if (u == NULL)
		return 0;
	return IS_OPER(u);
}

void s_del_user(char *u)
{
	irc_user *tmp;

	tmp = s_find_user(u);
	if (tmp == NULL)
		return;
	s_log(LOG_DEBUG, "Removing user %s.", tmp->nick);
	users = g_slist_remove(users, tmp);
	g_free(tmp);
	status.stats.global_users--;
}

irc_user *s_find_user(char *u)
{
	GSList *tmp;
	
	tmp = g_slist_find_custom(users, u, (GCompareFunc)s_usercmp);
	if (tmp == NULL) {
		s_log(LOG_DEBUG, "oopsie, didn't find it");
		return NULL;
	} else 
		return tmp->data;
}

GArray *s_find_users_by_host(char *u)
{
	GArray *usrs;
	struct hostsrc_args ha;
	
	usrs = g_array_new(TRUE, FALSE, sizeof(irc_user *));
	ha.resultstore = usrs;
	ha.searchstring = u;
	g_slist_foreach(users, (GFunc)s_hostsrc, &ha);
	return usrs;
}

int s_parse_userflags(char *modestr, int oldmode)
{
	int mode, direction;

	if (modestr == NULL) {
		s_log(LOG_DEBUG, "NULL mode?!");
		return 0;
	}
	mode = oldmode;
	direction = 0;
	for (; *modestr; modestr++) {
		switch (*modestr) {
			case '+':
				direction = 1;
				break;
			case '-':
				direction = 0;
				break;
			case 'r':
				mode = (direction) ? mode |  FLAG_REGISTERED:
						     mode & ~FLAG_REGISTERED;
				break;
			case 'o':
				mode = (direction) ? mode |  FLAG_OPERATOR:
					             mode & ~FLAG_OPERATOR;
				break;
			case 'a':
				mode = (direction) ? mode |  FLAG_SERVICESADMIN:
					             mode & ~FLAG_SERVICESADMIN;
				break;
		}
	}
	return mode;
}

void s_parse_mode_change(char *user, char *modes)
{
	irc_user *u;

	printf("parsing modes %s for user %s\n", modes, user);
	u = s_find_user(user);
	if (u == NULL)
		return;
	u->user_flags = s_parse_userflags((modes[0] == ':')
					  ? modes + 1
					  : modes, u->user_flags);
}

void s_parse_new_user(irc_msg *msg)
{
	char *gcos, *p;
	int nexttok;

	nexttok = 0;
	while (msg->msgtoks[nexttok] != NULL)
		nexttok++;
	if (nexttok < 10)
		/* invalid NICK */
		return;

	/* this, slightly obfuscated code, stores the user's
	 * GCOS to a variable, and points p to the beginning
	 * of the actual GCOS (skipping over ':'). */
	gcos = g_strjoinv(" ", msg->msgtoks + 9);
	p = (gcos[0] == ':') ? gcos + 1 : gcos;

	s_add_user(msg->msgtoks[0], msg->msgtoks[4], msg->msgtoks[5],
		   p, s_parse_userflags(msg->msgtoks[3], 0));
	g_free(gcos);
}

/* assumes irc_user is valid and allocated.
 * this function is bit weird ;).
 * EXPENSIVE FUNCTION */
irc_user *s_fill_user_by_nick(char *nick, irc_user *user)
{
	irc_user *u;
	
	if ((u = s_find_user(nick)) == NULL)
		return NULL;
        g_strlcpy(user->nick, u->nick, MAXNICKLEN);
        g_strlcpy(user->username, u->username, MAXUSERLEN);
        g_strlcpy(user->host, u->host, MAXHOSTLEN);
        g_strlcpy(user->gcos, u->gcos, MAXREALLEN);
	user->user_flags = u->user_flags;
	return user;
}
